import re
from dataclasses import dataclass, field
from typing import Tuple, Dict, Any, Iterable, List, Collection, Type

from data_to_paper.env import JSON_MODE, WRITING_MODEL_ENGINE
from data_to_paper.servers.model_engine import ModelEngine
from data_to_paper.text import dedent_triple_quote_str, word_count
from data_to_paper.base_steps.result_converser import Rewind
from data_to_paper.base_steps import BaseProductsQuotedReviewGPT, PythonDictReviewBackgroundProductsConverser, \
    PythonDictWithDefinedKeysReviewBackgroundProductsConverser
from data_to_paper.base_products.product import ValueProduct

from data_to_paper.servers.custom_types import Citation

from .cast import ScientificAgent
from .product_types import GoalAndHypothesisProduct, MostSimilarPapersProduct, NoveltyAssessmentProduct, \
    HypothesisTestingPlanProduct
from .scientific_products import ScientificProducts
from .writing_steps import ShowCitationProducts


@dataclass
class ScientificProductsQuotedReviewGPT(BaseProductsQuotedReviewGPT):

    sentence_to_add_at_the_end_of_performer_response: str = dedent_triple_quote_str("""
        Please provide feedback on the above {goal_noun}, with specific attention to whether it can be \t
        studied using only the provided dataset, without requiring any additional data \t
        (pay attention to using only data explicitly available in the provided headers of our data files \t
        as described in our dataset, above).
        Do not suggest changes to the {goal_noun} that may require data not available in our dataset.
        If you are satisfied, respond with "{termination_phrase}".
        """)


@dataclass
class GoalReviewGPT(ScientificProductsQuotedReviewGPT):
    model_engine: ModelEngine = WRITING_MODEL_ENGINE
    LLM_PARAMETERS = {'temperature': 1.0}
    max_reviewing_rounds: int = 1
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers',
                                                  'codes_and_outputs:data_exploration')
    conversation_name: str = 'Research Goal'
    other_conversation_name: str = 'research_goal_reviewer'
    goal_noun: str = 'research goal and hypothesis'
    goal_verb: str = 'suggest'
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.GoalReviewer
    termination_phrase: str = \
        'The research goal does not require any changes'

    project_specific_goal_guidelines: str = ""
    goal_guidelines: str = dedent_triple_quote_str("""\n
        Guidelines:

        * Try to avoid trivial hypotheses (like just testing for simple linear associations).
        Instead, you could perhaps explore more complex associations and relationships, like testing for \t
        moderation effects or interactions between variables.

        * Do not limit yourself to the provided data structure and variables; \t
        you can create new variables from the existing ones, and use them in your hypotheses.

        * Make sure that your suggested hypothesis can be studied using only the provided dataset, \t
        without requiring any additional data. In particular, pay attention to using only data available \t
        based on the provided headers of our data files (see "{data_file_descriptions_no_headers}", above).

        {project_specific_goal_guidelines}\t
        * Do not suggest methodology. Just the goal and an hypothesis. 
        """)
    mission_prompt: str = dedent_triple_quote_str("""\n
        Please suggest a research goal and an hypothesis that can be studied using only the provided dataset. 
        The goal and hypothesis should be interesting and novel.
        {goal_guidelines}

        Your response should be formatted as {your_response_should_be_formatted_as}
        """)
    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a triple-backtick block, like this:
        ```
        # Research Goal: 
        <your research goal here>

        # Hypothesis: 
        <your hypothesis here>
        ```
        """)
    other_system_prompt: str = dedent_triple_quote_str("""
        You are a {reviewer} for a {performer} who needs to {goal_verb} {goal_noun}.
        """)
    sentence_to_add_at_the_end_of_performer_response: str = dedent_triple_quote_str("""

        Please provide constructive bullet-point feedback on the above {goal_noun}.

        Specifically: 
        * If the hypothesis cannot be tested using only the provided dataset (without \t
        requiring additional data), suggest how to modify the hypothesis to better fit the dataset.
        * If the hypothesis is not interesting and novel, suggest how to modify it to make it more interesting.
        * If the hypothesis is broad or convoluted, suggest how best to focus it on a single well defined question.


        Do not provide positive feedback; if these conditions are all satisfied, just respond with: 
        "{termination_phrase}".
        If you feel that the initial goal and hypothesis satisfy the above conditions, \t
        respond solely with "{termination_phrase}".
    """)

    product_type: Type[ValueProduct] = GoalAndHypothesisProduct

    def _check_extracted_text_and_update_valid_result(self, extracted_text: str):
        if '\n# Research Goal:' not in extracted_text or '\n# Hypothesis:' not in extracted_text:
            self._raise_self_response_error(
                title='# Incorrect response format',
                error_message='Your response should contain both a "# Research Goal:" and a "# Hypothesis:" section.',
            )
        self._update_valid_result(extracted_text)


@dataclass
class GetMostSimilarCitations(ShowCitationProducts, PythonDictReviewBackgroundProductsConverser):
    json_mode: bool = JSON_MODE
    products: ScientificProducts = None
    allow_citations_from_step: str = 'goal'
    max_reviewing_rounds: int = 0

    value_type: type = Dict[str, str]
    goal_noun: str = 'most similar papers'
    goal_verb: str = 'find'
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.GoalReviewer
    conversation_name: str = 'Identify Similar Publications'
    is_new_conversation: bool = None  # this will create "similar_citations_0", etc.
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers', 'research_goal',
                                                  'literature_search:goal:dataset', 'literature_search:goal:questions')

    mission_prompt: str = dedent_triple_quote_str("""
        From the literature search above, list up to 5 key papers whose results are most \t
        similar/overlapping with our research goal and hypothesis.

        Return your response as a {python_or_json} Dict[str, str], where the keys are bibtex ids of the papers, \t
        and the values are the titles of the papers. For example:

        ```{python_or_json}
        {
            "Smith2020TheAB": 
                "A title of a paper most overlapping with our goal and hypothesis",
            "Jones2021AssortedCD": 
                "Another title of a paper that is similar to our goal and hypothesis",
        }
        ```
    """)

    def _check_response_value(self, response_value: Any) -> Any:
        response_value = super()._check_response_value(response_value)
        available_citations = self._get_available_citations()
        bibtex_ids_to_citations: Dict[str, Citation] = \
            {citation.bibtex_id: citation for citation in available_citations}
        non_matching_ids = [key for key in response_value.keys() if key not in bibtex_ids_to_citations]
        if non_matching_ids:
            self._raise_self_response_error(
                title='Invalid bibtex ids',
                error_message=f'Invalid bibtex ids:\n{non_matching_ids}')

        # replace with correct citation titles
        response_value = type(response_value)({key: bibtex_ids_to_citations[key].title for key in response_value})
        return response_value

    def _get_overlapping_citations(self, ids_to_titles: Dict[str, str]) -> List[Citation]:
        available_citations = self._get_available_citations()
        return [citation for citation in available_citations if citation.bibtex_id in ids_to_titles]

    def _convert_valid_result_to_product(self, valid_result: Dict[str, str]) -> MostSimilarPapersProduct:
        return MostSimilarPapersProduct(value=self._get_overlapping_citations(valid_result))

    def _convert_product_back_to_valid_result(self, product: MostSimilarPapersProduct) -> Dict[str, str]:
        return {citation.bibtex_id: citation.title for citation in product}


@dataclass
class NoveltyAssessmentReview(ShowCitationProducts, PythonDictWithDefinedKeysReviewBackgroundProductsConverser):
    json_mode: bool = JSON_MODE
    products: ScientificProducts = None
    value_type: type = Dict[str, Any]
    allowed_values_for_keys: Dict[str, Iterable] = field(default_factory=lambda: {'choice': ('OK', 'REVISE')})
    default_rewind_for_result_error: Rewind = Rewind.AS_FRESH_CORRECTION  # to maintain chain of thought
    goal_noun: str = 'novelty assessment'
    goal_verb: str = 'check'
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.GoalReviewer
    conversation_name: str = 'Assess Goal Novelty'
    requested_keys: Collection[str] = ('similarities', 'differences', 'choice', 'explanation')
    is_new_conversation: bool = None  # this will create "research_goal_0", etc.
    background_product_fields: Tuple[str, ...] = ('general_dataset_description', 'research_goal',
                                                  'most_similar_papers')
    sentence_to_add_at_the_end_of_reviewer_response: str = dedent_triple_quote_str("""
        Please correct your {goal_noun} based on the feedback provided.
        Make sure to return your full assessment, as {your_response_should_be_formatted_as}.
        """)

    mission_prompt: str = dedent_triple_quote_str("""
        We would like to assess the novelty of our {research_goal} with respect to the literature.
        Given the related papers listed above, please return \t
        {your_response_should_be_formatted_as}. Where:

        * "similarities": Provide a List[str] of potential similarities between our goal and hypothesis, \t
        and the related papers listed above.

        * "differences": Provide a List[str] of potential differences, if any, between our stated {research_goal} \t
        and the related papers listed above.

        * "choice": Given your assessment above, choose one of the following two options:

        a. Our goal and hypothesis offer a significant novelty compared to existing literature, and \t
        will likely lead to interesting and novel findings {"choice": "OK"}.

        b. Our goal and hypothesis have overlap with existing literature, and I can suggest ways to \t
        revise them to make them more novel {"choice": "REVISE"}.

        * "explanation": Provide a brief explanation of your choice.

        For example:

        ```{python_or_json}
        {
            "similarities": [
                "Our research goal is similar to the paper by ... in that ...",
                "Our research goal somewhat overlaps with the findings of ...",
                "Our hypothesis is similar to the paper by ... in that ..."
            ],
            "differences": [
                "Our goal and hypothesis are distinct because ...",
                "Our hypothesis differs from the paper by ... in that ..."
            ],

            "choice": "OK",
            "explanation": "While our goal and hypothesis have some overlap with existing literature, \t
        I believe that the ... aspect of our research is novel and will lead to ..."

            # or:

            "choice": "REVISE"
            "explanation": "The overlap with the result of ... is too significant, and I think we can \t
        revise our goal to make it more novel, for example by ..."
        }
        ```
        """)

    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a {python_or_json} dictionary, like this:"
        {"similarities": List[str], "differences": List[str], "choice": str, "explanation": str}
        """)

    product_type: Type[ValueProduct] = NoveltyAssessmentProduct

    def _check_response_value(self, response_value: Any) -> Any:
        response_value = super()._check_response_value(response_value)
        errors = []
        if response_value["choice"] not in ["OK", "REVISE"]:
            errors.append(f'Invalid choice: {response_value["choice"]}. Choose "OK" or "REVISE".')
        if not isinstance(response_value["explanation"], str):
            errors.append(f"Explanation must be a string.")
        for key in ["similarities", "differences"]:
            if not isinstance(response_value[key], list):
                errors.append(f'"{key}" must be a list of strings.')
            for item in response_value[key]:
                if not isinstance(item, str):
                    errors.append(f'Each item in "{key}" must be a string.')
        if errors:
            errors = '\n'.join(errors)
            self._raise_self_response_error(
                title='# Errors in response structure',
                error_message=errors,
            )
        return response_value


@dataclass
class ReGoalReviewGPT(GoalReviewGPT):
    is_new_conversation: bool = None
    max_reviewing_rounds: int = 0
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers',
                                                  'codes_and_outputs:data_exploration',
                                                  'research_goal', 'most_similar_papers')
    mission_prompt: str = dedent_triple_quote_str("""
        Based on the result of the literature search above, \t
        please revise, or completely re-write, the research goal and hypothesis that we have so that they \t
        do not completely overlap existing literature.
        {goal_guidelines}

        Your response should be formatted as {your_response_should_be_formatted_as}
        """)


@dataclass
class HypothesesTestingPlanReviewGPT(PythonDictWithDefinedKeysReviewBackgroundProductsConverser):
    json_mode: bool = JSON_MODE
    value_type: type = Dict[str, Dict[str, str]]
    requested_keys: Collection[str] = ('ISSUES', 'HYPOTHESES')
    max_valid_response_iterations: int = 4
    max_hypothesis_count: int = 3
    max_reviewing_rounds: int = 0  # 0 for no review cycles
    default_rewind_for_result_error: Rewind = Rewind.AS_FRESH_CORRECTION  # to maintain chain of thought
    background_product_fields: Tuple[str, ...] = ('data_file_descriptions_no_headers',
                                                  'codes_and_outputs:data_exploration',
                                                  'research_goal')
    your_response_should_be_formatted_as: str = dedent_triple_quote_str("""
        a {python_or_json} dictionary with the following structure:
        ```{python_or_json}
        {
            "ISSUES": {
                "<Issue>": "<Description of the issue and how it should be accounted for>",
                "<Another issue>": "...",
                "etc": "..."
            }

            "HYPOTHESES": {
                "<Hypothesis>": "<Statistical test>",
                "<another Hypothesis>": "...",
                "etc": "..."
            }
        }
        ```
        """)
    conversation_name: str = 'Hypothesis Testing Plan'
    is_new_conversation: bool = None  # this will create "hyp_testing_plan_0", etc.
    goal_noun: str = 'hypothesis testing plan'
    goal_verb: str = 'write'
    mission_prompt: str = dedent_triple_quote_str("""
        We would like to test the specified hypotheses using the provided dataset.
        We need to think of the relevant statistical issues and the most adequate statistical tests for each hypothesis.

        Review the "{data_file_descriptions_no_headers}" and "{codes_and_outputs:data_exploration}" provided above, \t
        and return your assessment as {your_response_should_be_formatted_as}

        - ISSUES:
        The keys of this dictionary should briefly describe the statistical issues that we should account for.
        The values should describe the issue and how it should be accounted for in the statistical tests.
        For possible issues (keys), consider for example:
        * Confounding variables (see available variables in the dataset that we can adjust for).
        * Missing data points.
        * Dependencies between data points.
        * Multiple comparisons.
        * Imbalanced data.
        * Any other relevant statistical issues.

        - HYPOTHESES
        The keys of this dictionary should briefly describe each of our hypotheses.
        The values of this dictionary \t
        should specify the most adequate statistical test for each hypothesis, \t
        and describe how it should be performed while accounting for any issues you have outlined above.

        For each of our hypotheses, suggest a *single* statistical test.
        If there are several possible ways to test a given hypothesis, specify only *one* statistical test \t
        (the simplest one).

        Example:

        ```{python_or_json}
        {
            "ISSUES": {
                "Missing data points": 
                    "Based on the {codes_and_outputs:data_exploration}, \t
        we should drop lines with missing data in ...",
                "Confounding variables": 
                    "We should adjust for ..."
            },
            "HYPOTHESES": {
                "xxx is associated with yyy and zzz":
                    "Linear regression with xxx as the independent variable and \t
        yyy and zzz as the dependent variables while adjusting for aaa, bbb, ccc",
                "The association between xxx and yyy is moderated by zzz": 
                    "Repeat the above linear regression, while adding the interaction term between yyy and zzz"
        }
        ```

        These of course are just examples. Your actual response should be based on the \t
        "{research_goal}", "{data_file_descriptions_no_headers}", and "{codes_and_outputs:data_exploration}".

        Note how in the example shown the different hypotheses are connected to each other, building towards a single
        study goal.

        Remember to return a valid {python_or_json} dictionary with the structure described above.
        """)
    assistant_agent: ScientificAgent = ScientificAgent.Performer
    user_agent: ScientificAgent = ScientificAgent.PlanReviewer
    product_type: Type[ValueProduct] = HypothesisTestingPlanProduct

    def _check_response_value(self, response_value: Any) -> Any:
        """
        Strip "hypothesis x" from the keys of the response value.
        """
        response_value = super()._check_response_value(response_value)

        # what we have is "usable" at this point:
        self._update_valid_result(response_value)

        hypotheses = response_value['HYPOTHESES']

        if len(hypotheses) > self.max_hypothesis_count:
            self._raise_self_response_error(
                title='# Too many hypotheses',
                error_message=dedent_triple_quote_str(f"""
                    Please do not specify more than {self.max_hypothesis_count} hypotheses.
                    Revise your response to return a maximum of {self.max_hypothesis_count} hypotheses, \t
                    which should all build towards a single study goal.
                    """)
            )

        if not hypotheses:
            self._raise_self_response_error(
                title='# No hypotheses',
                error_message='Please specify at least one hypothesis.'
            )

        # We want explicit hypotheses. Check that the hypotheses have at least 7 words:
        for hypothesis, test in hypotheses.items():
            if word_count(hypothesis) < 5:
                self._raise_self_response_error(
                    title='# Hypothesis too short',
                    error_message=f'The hypothesis "{hypothesis}" is too short.\n'
                                  f'Please provide a more explicit and specific hypothesis description.'
                )
            if word_count(test) < 12:
                self._raise_self_response_error(
                    title='# Hypothesis test too short',
                    error_message=f'The test description "{test}" is too short.\n'
                                  f'Please provide a more explicit and specific test description.'
                )

        # remove "hypothesis x" from the keys (we don't want the hypotheses to be numbered)
        hypotheses = type(hypotheses)(
            {re.sub(pattern=r'hypothesis \d+:|hypothesis:|hypothesis :',
                    repl='', string=k, flags=re.IGNORECASE).strip(): v
             for k, v in hypotheses.items()})
        response_value['HYPOTHESES'] = hypotheses
        return response_value



import os
import shutil
import tempfile
import uuid
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Union, Iterable
from fnmatch import fnmatch


def is_valid_filename(filename):
    # Regular expression for validating the filename
    pattern = r'^[a-zA-Z0-9_-]+$'
    # Match the pattern with the filename
    if re.match(pattern, filename):
        return True
    else:
        return False


def clear_directory(directory: Union[Path, str], create_if_missing: bool = True):
    """
    Clear the directory of all files and subdirectories.
    """
    directory = Path(directory)
    if not directory.exists():
        if create_if_missing:
            directory.mkdir(parents=True)
        else:
            raise FileNotFoundError(f'Directory {directory} does not exist.')
    for item in directory.iterdir():
        if item.is_dir():
            shutil.rmtree(item)
        else:
            item.unlink()


def is_name_matches_list_of_wildcard_names(file_name: str, list_of_filenames: Iterable[str]):
    """
    Check if file_name matches any of the wildcard filenames in list_of_filenames.
    Wildcard filenames are filenames that end with '*'.
    The wildcard '*' can be in the beginning or in the end of the filename.
    """
    for wildcard_filename in list_of_filenames:
        if fnmatch(file_name, wildcard_filename):
            return True
    return False


@contextmanager
def run_in_temp_directory():
    """
    Run code in a temporary folder.
    The folder is deleted after the code is done running.
    """
    cwd = os.getcwd()
    folder = os.path.join(tempfile.gettempdir(), f'data_to_paper_temp_{uuid.uuid4()}')
    if not os.path.exists(folder):
        os.mkdir(folder)
    os.chdir(folder)
    try:
        yield folder
    finally:
        os.chdir(cwd)
        shutil.rmtree(folder)


# context manager to run in a given directory:
@contextmanager
def run_in_directory(folder: Union[Path, str] = None) -> Union[Path, str]:
    """
    Run code in a specific folder.
    If folder is None, run in the current folder.
    """
    cwd = os.getcwd()
    if folder is not None:
        os.chdir(folder)
    try:
        yield folder
    finally:
        os.chdir(cwd)


def get_non_existing_file_name(file_path: Union[Path, str]) -> Union[Path, str]:
    """
    If file_path already exists, add a number to the end of the file name.
    """
    file_path = Path(file_path)
    i = 0
    while True:
        new_file_path = file_path.with_name(f'{file_path.stem}_{i}{file_path.suffix}')
        if not new_file_path.exists():
            return new_file_path
        i += 1



from dataclasses import dataclass
from typing import Optional, Any

from data_to_paper.run_gpt_code.base_run_contexts import RegisteredRunContext


@dataclass
class SetRandomSeeds(RegisteredRunContext):
    """
    Set the global seed of:
    - numpy (np.random.seed)
    - random (random.seed)
    """
    random_seed: Optional[int] = 0  # None to disable
    _np_seed: Optional[Any] = None
    _random_seed: Optional[Any] = None

    def __enter__(self):
        if self.random_seed is not None:
            import numpy as np
            import random
            self._np_seed = np.random.get_state()
            self._random_seed = random.getstate()
            np.random.seed(self.random_seed)
            random.seed(self.random_seed)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.random_seed is not None:
            import numpy as np
            import random
            np.random.set_state(self._np_seed)
            random.setstate(self._random_seed)
            self._np_seed = None
            self._random_seed = None
        return super().__exit__(exc_type, exc_val, exc_tb)



import pandas as pd
from pandas import DataFrame

from data_to_paper.utils.types import ListBasedSet


def _extract_from_index_or_multiindex(index_or_multiindex, with_title: bool = True, string_only: bool = False
                                      ) -> ListBasedSet:
    if isinstance(index_or_multiindex, pd.MultiIndex):
        result = ListBasedSet(item for tuple_level in index_or_multiindex for item in tuple_level)
    else:
        result = ListBasedSet(index_or_multiindex)
    if with_title:
        result.add(index_or_multiindex.name)
    if string_only:
        result = ListBasedSet(item for item in result if isinstance(item, str))
    return result


def extract_df_column_labels(df: DataFrame, with_title: bool = True, string_only: bool = False) -> ListBasedSet:
    return _extract_from_index_or_multiindex(df.columns, with_title=with_title, string_only=string_only)


def extract_df_row_labels(df: DataFrame, with_title: bool = True, string_only: bool = False) -> ListBasedSet:
    return _extract_from_index_or_multiindex(df.index, with_title=with_title, string_only=string_only)


def extract_df_axes_labels(df: DataFrame, index: bool = True, header: bool = True, with_title: bool = True,
                           string_only: bool = False) -> ListBasedSet:
    axes_labels = ListBasedSet()
    if header:
        axes_labels |= extract_df_column_labels(df, with_title=with_title, string_only=string_only)
    if index:
        axes_labels |= extract_df_row_labels(df, with_title=with_title, string_only=string_only)
    return axes_labels



from __future__ import annotations

import os
from abc import ABCMeta
from dataclasses import dataclass
from requests import Response
from typing import Union

from data_to_paper.terminate.exceptions import TerminateException
from data_to_paper.text import dedent_triple_quote_str


@dataclass
class APIKey:
    key: str
    key_name: str

    @classmethod
    def from_env(cls, key_name: str) -> APIKey:
        key = os.environ.get(key_name)
        return cls(key, key_name)


@dataclass
class BaseServerErrorException(TerminateException, metaclass=ABCMeta):
    server: str


@dataclass
class ServerErrorException(BaseServerErrorException):
    """
    Error raised server wasn't able to respond.
    """

    response: Union[Response, Exception]

    def _get_response_as_str(self):
        try:
            return f"Server status code: {self.response.status_code}\n{self.response.reason}"
        except AttributeError:
            return str(self.response)

    def __str__(self):
        return (
            f"Request to `{self.server}` server failed with the following error:\n```error\n"
            f"{self._get_response_as_str()}\n```"
        )


@dataclass
class BaseAPIKeyError(BaseServerErrorException, metaclass=ABCMeta):
    api_key: APIKey
    instructions: str = dedent_triple_quote_str("""
        To set up the API keys as environment variables on your system, see:
        https://help.openai.com/en/articles/5112595-best-practices-for-api-key-safety
        """)


class MissingAPIKeyError(BaseAPIKeyError):
    def __str__(self):
        return (
            f"The API key for `{self.server}` is missing.\n"
            f"You need to set the `{self.api_key.key_name}` environment variable.\n"
            f"\n{self.instructions}"
        )


@dataclass
class MissingSemanticScholarAPIKeyError(MissingAPIKeyError):
    def __str__(self):
        return (
            f"WARNING: SEMANTIC SCHOLAR API key is missing. The server should work without an API key, "
            f"but it will be slower and depend on the availability of the API endpoint.\n"
            f"Its recommended setting the `SEMANTIC_SCHOLAR_API_KEY` environment variable.\n"
            f"\n{self.instructions}"
        )

    def __repr__(self):
        return self.__str__()


@dataclass
class InvalidAPIKeyError(BaseAPIKeyError):
    response: Union[Response, Exception] = None

    def __str__(self):
        return (
            f"The API key for `{self.server}` is invalid.\n"
            f"Trying to connect to the server with the key:\n`{self.api_key.key}`\n"
            f"Connection failed with the following error:\n"
            f"---\n```error\n{self.response}\n```\n---\n"
            f"\n{self.instructions}"
        )



from dataclasses import dataclass, field
from typing import Iterable

from data_to_paper.utils.nice_list import NiceList

from .pvalue import TrackPValueCreationFuncs
from .random import SetRandomSeeds
from .scipy.override_scipy import ScipyPValueOverride
from .sklearn.override_sklearn import SklearnFitOverride, SklearnSearchLimitCheck, SklearnRandomStateOverride, \
    SklearnNNSizeOverride, SklearnPValue
from .statsmodels.override_statsmodels import StatsmodelsFitPValueOverride, StatsmodelsMultitestPValueOverride, \
    StatsmodelsAnovaPValueOverride, StatsmodelsMulticompPValueOverride
from ..base_run_contexts import RunContext, MultiRunContext
from ..run_issues import CodeProblem, RunIssue


@dataclass
class OverrideStatisticsPackages(MultiRunContext):
    """
    Base context manager for running GPT code.
    """
    issue_if_statistics_test_not_called: bool = True

    contexts: Iterable[RunContext] = field(default_factory=lambda: [
        ScipyPValueOverride(prevent_unpacking=True),
        SklearnFitOverride(),
        StatsmodelsFitPValueOverride(),
        StatsmodelsMultitestPValueOverride(),
        StatsmodelsAnovaPValueOverride(),
        StatsmodelsMulticompPValueOverride(),
        SklearnPValue(),
        SklearnSearchLimitCheck(),
        SklearnRandomStateOverride(),  # see comment below
        SklearnNNSizeOverride(),
        SetRandomSeeds(random_seed=0),
    ])
    # TODO: SklearnRandomStateOverride is kept for backwards compatability.
    #  It is likely not needed if we use SetRandomSeeds
    #  May undesirably lead to non-random results when iterating within the same run!

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.issue_if_statistics_test_not_called:
            stat_test_called = False
            stat_packages = set()
            for context in self.contexts:
                if isinstance(context, TrackPValueCreationFuncs):
                    stat_packages |= set(context.package_names)
                    if context.pvalue_creating_funcs:
                        stat_test_called = True
            if not stat_test_called:
                self.issues.append(RunIssue(
                    category='Statistics: good practice',
                    issue="We are writing code for an hypothesis-testing paper, "
                          "but your code does not call any statistical-testing function that returns a p-value.",
                    instructions="Please make sure that you perform a statistical-test with either "
                                 "{}.".format(sorted(NiceList(stat_packages, last_separator=', or '))),
                    code_problem=CodeProblem.NonBreakingRuntimeIssue,
                ))
        return super().__exit__(exc_type, exc_val, exc_tb)



from dataclasses import dataclass


@dataclass
class Copier:
    """
    Base class for dataclass classes that have specific attributes that should be copied from other instances of the
    same-class, sub-classes or super-classes.
    """
    COPY_ATTRIBUTES = ()

    @classmethod
    def get_copy_attributes(cls):
        """
        Recursively gather COPY_ATTRIBUTES from the current class and its superclasses,
        but only from classes that are subclasses of Copier.
        """
        # Collect attributes from current class and all parent classes that are subclasses of Copier
        attrs = set()
        for base in cls.__mro__:
            if issubclass(base, Copier):
                attrs.update(getattr(base, 'COPY_ATTRIBUTES'))
        return attrs

    @classmethod
    def from_(cls, other, **kwargs):
        """
        Create a new instance of the class, copying attributes from another instance of the same class.
        """
        k = dict()
        for attr in cls.get_copy_attributes():
            if attr in kwargs:
                k[attr] = kwargs.pop(attr)
            elif hasattr(other, attr):
                k[attr] = getattr(other, attr)
        return cls(**k, **kwargs)



from data_to_paper.interactive.base_app_startup import DataFilesStartDialog, TextEditWithHeader, SingleFileWidget


class ToyStartDialog(DataFilesStartDialog):
    def _set_style(self):
        super()._set_style()
        self.resize(600, 500)

    def _create_widgets(self):
        return {
            'research_goal': TextEditWithHeader(
                "Research Goal", "Specify a funny research goal.", on_change=self.update_start_button_state),
            'files_widget': SingleFileWidget(on_change=self.update_start_button_state),
        }

    def _convert_config_to_widgets(self, config):
        self.widgets['research_goal'].setPlainText(config.get('research_goal', '') or '')
        super()._convert_config_to_widgets(config)

    def _convert_widgets_to_config(self) -> dict:
        config = self.current_config
        config['general_description'] = ''
        config['research_goal'] = self.widgets['research_goal'].toPlainText() or None
        config = super()._convert_widgets_to_config()
        return config

    def _get_mandatory_items_to_start(self):
        research_goal_filled = bool(self.widgets['research_goal'].toPlainText().strip())
        files_filled = any(file_widget.abs_path and file_widget.description_edit.toPlainText().strip()
                           for file_widget in self._get_date_files_widget().get_all_file_widgets())

        return research_goal_filled, files_filled



from dataclasses import dataclass
from typing import Tuple, Optional, Dict, Any, Collection, List

from data_to_paper.base_steps.request_code import CodeReviewPrompt
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirements, \
    TextContentOutputFileRequirement, NumericTextContentOutputFileRequirement
from data_to_paper.research_types.hypothesis_testing.cast import ScientificAgent
from data_to_paper.research_types.hypothesis_testing.coding.base_code_conversers import BaseScientificCodeProductsGPT
from data_to_paper.research_types.hypothesis_testing.coding.utils import create_pandas_and_stats_contexts
from data_to_paper.run_gpt_code.run_issues import RunIssue, CodeProblem
from data_to_paper.text import dedent_triple_quote_str
from data_to_paper.text.highlighted_text import output_to_highlighted_html
from data_to_paper.utils.nice_list import NiceList


@dataclass(frozen=True)
class EnforceContentOutputFileRequirement(TextContentOutputFileRequirement, NumericTextContentOutputFileRequirement):
    should_keep_file: bool = False
    headers_required_in_output: Tuple[str, ...] = \
        ('# Data Size', '# Summary Statistics', '# Categorical Variables', '# Missing Values')

    def _get_content_and_header_for_app_html(self, content: Any, filename: str = None, num_file: int = 0,
                                             level: int = 3, view_purpose: ViewPurpose = ViewPurpose.APP_HTML):
        content = self._to_str(content, filename, num_file, view_purpose)
        content = output_to_highlighted_html(content)
        return content, f'<h{level}>{filename}</h{level}>'

    def get_issues_for_output_file_content(self, filename: str, content: str) -> List[RunIssue]:
        issues = super().get_issues_for_output_file_content(filename, content)
        if issues:
            return issues

        missing_headers = [header for header in self.headers_required_in_output if header not in content]
        if missing_headers:
            issues.append(RunIssue(
                category='Problem in output file(s)',
                item=filename,
                issue=f'The output file "{filename}" should have the following headers: '
                      f'{NiceList(self.headers_required_in_output, wrap_with="`")}.\n'
                      f'But, these headers are missing: '
                      f'{NiceList(missing_headers, wrap_with="`")}.',
                code_problem=CodeProblem.OutputFileCallingSyntax,
            ))

        return issues

    def _convert_content_to_labeled_text(self, content: Any, filename: str = None, num_file: int = 0,
                                         view_purpose: ViewPurpose = None) -> str:
        return content


@dataclass
class DataExplorationCodeProductsGPT(BaseScientificCodeProductsGPT):
    code_step: str = 'data_exploration'
    background_product_fields: Tuple[str, ...] = ('all_file_descriptions', )
    user_agent: ScientificAgent = ScientificAgent.DataExplorer
    allow_data_files_from_sections: Tuple[Optional[str]] = (None, )

    supported_packages: Tuple[str, ...] = ('pandas', 'numpy', 'scipy')

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements([EnforceContentOutputFileRequirement('data_exploration.txt')])

    @property
    def output_filename(self) -> str:
        return self.output_file_requirements[0].generic_filename

    mission_prompt: str = dedent_triple_quote_str("""
        As part of a data-exploration phase, please write a complete short Python code for getting a \t
        first sense of the data. 

        Your code should create an output text file named "{output_filename}", which should \t
        contain a summary of the data.

        The output file should be self-contained; any results you choose to save to this file \t
        should be accompanied with a short header.

        The output file should be formatted as follows:

        ```
        # Data Size
        <Measure of the scale of our data (e.g., number of rows, number of columns)>

        # Summary Statistics
        <Summary statistics of all or key variables>

        # Categorical Variables
        <As applicable, list here categorical values and their most common values>

        # Missing Values
        <Counts of missing, unknown, or undefined values>
        <As applicable, counts of special numeric values that stand for unknown/undefined if any \t
        (check in the "{all_file_descriptions}" above for any)>

        # <title of other summary you deem relevant, if any>
        <Add any other summary of the data you deem relevant>

        # <etc for any other summary you deem relevant.>
        ```

        If any of the above sections is not applicable, then write "# Not Applicable" under that section.

        If needed, you can use the following packages which are already installed:
        {supported_packages}

        Do not provide a sketch or pseudocode; write a complete runnable code.
        Do not create any graphics, figures or any plots.
        Do not send any presumed output examples.
        """)

    code_review_prompts: Collection[CodeReviewPrompt] = (
        CodeReviewPrompt('output file', '*', False, dedent_triple_quote_str("""
        I ran your code.

        Here is the content of the output file that the code created:

        {file_contents_str}

        Please carefully check the Python code and the output for possible issues, and \t
        provide a point-by-point assessment. 

        {code_review_formatting_instructions}

        For example:
        ```{python_or_json}
        {
            "NaN values in the output file":
                ["CONCERN", "The output contains NaN values in ..."],
            "Output file should be self-contained":
                ["CONCERN", "A header is missing for ..."],
            "Output file should contain all the required analysis": 
                ["OK", "Nothing is missing"],
            "Sensible results": 
                ["CONCERN", "The average of ... does not make sense"],
            "<Any other issues you find>":
                ["CONCERN", "<Issue description>"],
            "<Any other point you checked and asserted is OK>":
                ["OK", "<Assertion description>"],
        }
        ```

        {code_review_notes}
        """)),
    )

    def _get_additional_contexts(self) -> Optional[Dict[str, Any]]:
        return create_pandas_and_stats_contexts(allow_dataframes_to_change_existing_series=False,
                                                enforce_saving_altered_dataframes=False)



from dataclasses import dataclass
from typing import Tuple, Optional, Dict, List, Union

from matplotlib import pyplot as plt
from matplotlib.transforms import BboxBase


def get_xy_coordinates_of_df_plot(df, x=None, y=None, kind='line') -> Dict[int, Dict[int, Tuple[float, float]]]:
    """
    Plots the DataFrame and retrieves x and y coordinates for each data point.
    Returns a dictionary of dictionaries, where the first key is the column index and the second key is the row index.
    """
    # Create the plot
    ax = df.plot(x=x, y=y, kind=kind, legend=False)

    coords = {}
    if kind == 'bar':
        # Handle bar plots
        for col_index, container in enumerate(ax.containers):
            coords[col_index] = {}
            for rect, row_index in zip(container, range(len(df))):
                coords[col_index][row_index] = (rect.get_x() + rect.get_width() / 2, rect.get_height())
    else:
        # Handle line and other plots
        for line, col_index in zip(ax.lines, range(len(df.columns))):
            coords[col_index] = {}
            for x_val, y_val, row_index in zip(line.get_xdata(), line.get_ydata(), range(len(df))):
                coords[col_index][row_index] = (x_val, y_val)

    plt.close()  # Close the plot to avoid display
    return coords


def are_axes_numeric(ax: plt.Axes) -> Tuple[bool, bool]:
    """
    For each axes, x and y, check if all the ticks are numeric.
    """
    def is_numeric(tick):
        tick = tick.get_text()
        if isinstance(tick, (int, float)):
            return True
        try:
            float(tick)
            return True
        except ValueError:
            return False

    return all(is_numeric(tick) for tick in ax.get_xticklabels()), \
        all(is_numeric(tick) for tick in ax.get_yticklabels())


def replace_singleton_legend_with_axis_label(ax: plt.Axes, kind: str) -> Optional[str]:
    """
    Replace a singleton legend with an axis label.
    """
    # suppress all warnings that are not errors:
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        legend = ax.legend(handlelength=0.75, handleheight=0.75, handletextpad=0.5, borderpad=.25, borderaxespad=0.,
                           labelspacing=0.3, framealpha=0, bbox_to_anchor=(1.05, 0.5), loc='center left',
                           bbox_transform=ax.transAxes)
    for text in legend.get_texts():
        text.set_fontsize(9)
    legend_keys = [text.get_text() for text in legend.get_texts()]
    is_singleton = len(legend_keys) == 1
    if not is_singleton:
        return
    singleton_legend_key = legend_keys[0]
    if kind == 'barh':
        # Horizontal bar plot. The legend is the x-axis label.
        if ax.get_xlabel() == '':
            ax.set_xlabel(singleton_legend_key)
    else:
        # Normal plot. The legend is the y-axis label.
        if ax.get_ylabel() == '':
            ax.set_ylabel(singleton_legend_key)
    legend.remove()
    return singleton_legend_key


def add_grid_line_at_base_if_needed(ax: plt.Axes, h_or_v: str, color: str = 'black', linewidth: int = 1,
                                    linestyle: str = '--'):
    """
    Add a grid line at zero, if zero is within the axis limits.
    h_or_v: 'h' for horizontal, 'v' for vertical, 'hv' for both.
    """
    if 'h' in h_or_v:
        base_value = 1 if ax.get_yscale() == 'log' else 0
        if ax.get_ylim()[0] < base_value < ax.get_ylim()[1]:
            ax.axhline(base_value, color=color, linewidth=linewidth, linestyle=linestyle)
    if 'v' in h_or_v:
        base_value = 1 if ax.get_xscale() == 'log' else 0
        if ax.get_xlim()[0] < base_value < ax.get_xlim()[1]:
            ax.axvline(base_value, color=color, linewidth=linewidth, linestyle=linestyle)


def rotate_xticklabels_if_not_numeric(ax: plt.Axes):
    """
    Rotate the x-tick labels if they are not numeric.
    """
    x_numeric, _ = are_axes_numeric(ax)
    for label in ax.get_xticklabels():
        if not x_numeric:
            label.set_rotation(45)
            label.set_horizontalalignment('right')
            label.set_rotation_mode('anchor')
        else:
            label.set_rotation(0)


@dataclass(frozen=True)
class AxisParameters:
    title: str
    xlabel: str
    ylabel: str
    xscale: str
    yscale: str
    xlim: Tuple[float, float]
    ylim: Tuple[float, float]
    xticks: List[float]
    yticks: List[float]
    xticklabels: List[Union[str, float]]
    yticklabels: List[Union[str, float]]

    def is_x_axis_numeric(self) -> bool:
        """
        Check if all the ticks are numeric.
        """
        return all(isinstance(tick, (int, float)) for tick in self.xticks)

    def is_y_axis_numeric(self) -> bool:
        """
        Check if all the ticks are numeric.
        """
        return all(isinstance(tick, (int, float)) for tick in self.yticks)


def get_axis_parameters(ax: plt.Axes) -> AxisParameters:
    """
    Get the parameters of the axes.
    """
    return AxisParameters(
        title=ax.get_title(),
        xlabel=ax.get_xlabel(),
        ylabel=ax.get_ylabel(),
        xscale=ax.get_xscale(),
        yscale=ax.get_yscale(),
        xlim=ax.get_xlim(),
        ylim=ax.get_ylim(),
        xticks=ax.get_xticks(),
        yticks=ax.get_yticks(),
        xticklabels=ax.get_xticklabels(),
        yticklabels=ax.get_yticklabels(),
    )


def fit_fig_to_axes(fig, margin_pixels: int = 10, fit_height: bool = True, fit_width: bool = True):
    """
    Adjusts the figure size to snappily fit the axes, while keeping the axes physical size constant.
    fig: The figure to adjust.
    margin_pixels: The margin in pixels to add around the axes.
    fit_height: Whether to fit the height of the figure to the axes.
    fit_width: Whether to fit the width of the figure to the axes.
    """
    axs = fig.axes
    fig_width, fig_height = fig.get_size_inches()

    # get a tight bounding box around all axes
    renderer = fig.canvas.get_renderer()
    bbox_tight_all_pixels = BboxBase.union([ax.get_tightbbox(renderer) for ax in axs])

    bbox_tight_all = bbox_tight_all_pixels.padded(margin_pixels).transformed(fig.transFigure.inverted())

    # calculate the new figure size
    if fit_width:
        fig_width = bbox_tight_all.width * fig_width
    if fit_height:
        fig_height = bbox_tight_all.height * fig_height

    # set the new figure size
    fig.set_size_inches(fig_width, fig_height)

    # adjust the position of all axes to match the new figure size
    for ax in fig.axes:
        bbox = ax.get_position()
        if fit_width:
            left = (bbox.x0 - bbox_tight_all.x0) / bbox_tight_all.width
            width = bbox.width / bbox_tight_all.width
        else:
            left = bbox.x0
            width = bbox.width
        if fit_height:
            bottom = (bbox.y0 - bbox_tight_all.y0) / bbox_tight_all.height
            height = bbox.height / bbox_tight_all.height
        else:
            bottom = bbox.y0
            height = bbox.height
        ax.set_position([left, bottom, width, height])



import os
import time

import numpy as np

from typing import Dict, Union, List, Tuple

from data_to_paper.run_gpt_code.attr_replacers import AttrReplacer
from data_to_paper.run_gpt_code.overrides.pvalue import PValue, OnStrPValue, OnStr
from data_to_paper.run_gpt_code.run_issues import CodeProblem, RunIssue
from data_to_paper.utils.check_type import validate_value_type, WrongTypeException

BasicTypes = Union[str, int, float, bool, np.number, None, PValue]

AllowedType = Dict[str, Union[BasicTypes, Dict[str, BasicTypes], List[BasicTypes], Tuple[BasicTypes, ...]]]


def _pickle_dump_with_checks(obj, file, *args, original_func=None, context_manager: AttrReplacer = None, **kwargs):
    """
    Save a Dict[str, Any] to a pickle file.
    Check for content issues.
    """
    filename = file.name
    category = 'Use of `pickle.dump`'

    def close_and_remove_and_raise(e: Exception):
        file.close()
        time.sleep(0.1)
        os.remove(filename)  # the file is already open and has size 0
        raise e

    # Do not allow any other arguments
    if args or kwargs:
        close_and_remove_and_raise(RunIssue.from_current_tb(
            category=category,
            item=filename,
            issue="Please use `dump(obj, file)` with only the `obj` and `file` arguments.",
            instructions="Please do not specify any other arguments.",
            code_problem=CodeProblem.RuntimeError,
        ))

    # Only allow a dictionary with string keys and values that are basic types.
    if not isinstance(obj, dict) or not all(isinstance(key, str) for key in obj.keys()):
        close_and_remove_and_raise(RunIssue.from_current_tb(
            category=category,
            item=filename,
            issue="You can only save a dictionary with string keys and values that are basic types.",
            code_problem=CodeProblem.RuntimeError,
        ))

    try:
        validate_value_type(obj, AllowedType)
    except WrongTypeException as e:
        close_and_remove_and_raise(RunIssue.from_current_tb(
            category=category,
            item=filename,
            issue="You can only save a dictionary with string keys and values that are basic types.\n"
                  "Checking the type of the saved object, I found: " + e.get_found_description(),
            instructions=f"Please make sure that the object you are saving to the {filename} is a dictionary with "
                         f"string keys and simple-type values.",
            code_problem=CodeProblem.RuntimeError,
        ))

    # try str of the dict. This will raise an error if the dict contains object that
    #  we do not allow to show as text.
    try:
        with OnStrPValue(OnStr.SMALLER_THAN):
            s = str(obj)
    except Exception as e:
        close_and_remove_and_raise(RunIssue.from_current_tb(
            category=category,
            item=filename,
            issue="The saved dictionary contains objects that cannot be shown as text.\n"
                  "It created the following error:\n" + str(e),
            code_problem=CodeProblem.RuntimeError,
        ))

    if len(s) > 500:
        close_and_remove_and_raise(RunIssue.from_current_tb(
            category=category,
            item=filename,
            issue="The saved dictionary is too large to be shown as text.",
            instructions="Please reduce the amount of data you are saving to the file.",
            code_problem=CodeProblem.RuntimeError,
        ))

    with PValue.BEHAVE_NORMALLY.temporary_set(True):
        return original_func(obj, file)


def get_pickle_dump_attr_replacer():
    return AttrReplacer(obj_import_str='pickle', attr='dump', wrapper=_pickle_dump_with_checks,
                        send_context_to_wrapper=True, send_original_to_wrapper=True)



from .df_to_figure import df_to_figure
from .df_to_latex import df_to_latex

from .consts import DF_ALLOWED_VALUE_TYPES, ALLOWED_PLOT_KINDS
from .describe import describe_df, describe_value

assert df_to_figure.__name__ == 'df_to_figure'
assert df_to_latex.__name__ == 'df_to_latex'



from typing import Dict, Any, Optional

from data_to_paper.run_gpt_code.overrides.contexts import OverrideStatisticsPackages
from data_to_paper.run_gpt_code.overrides.dataframes import TrackDataFrames


def create_pandas_and_stats_contexts(allow_dataframes_to_change_existing_series: Optional[bool] = False,
                                     enforce_saving_altered_dataframes: bool = False,
                                     issue_if_statistics_test_not_called: bool = False,
                                     ) -> Dict[str, Any]:
    return {
        'TrackDataFrames': TrackDataFrames(
            allow_dataframes_to_change_existing_series=allow_dataframes_to_change_existing_series,
            enforce_saving_altered_dataframes=enforce_saving_altered_dataframes,
        ),
        'OverrideStatisticsPackages': OverrideStatisticsPackages(
            issue_if_statistics_test_not_called=issue_if_statistics_test_not_called),
    }



from typing import Union, Tuple, Optional

StrOrTupleStr = Union[str, Tuple[str, str]]


def nicely_join(words: list, wrap_with: StrOrTupleStr = '',
                prefix: StrOrTupleStr = '', suffix: StrOrTupleStr = '',
                separator: str = ', ', last_separator: Optional[str] = None, empty_str: str = ''):
    """
    Concatenate a list of words with commas and an 'and' at the end.

    wrap_with: if str: wrap each word with the provided string. if tuple: wrap each word with the first string
    on the left and the second string on the right.

    prefix: if str: add the provided string before the concatenated words. if tuple the first string is for singular
    and the second string is for plural.

    suffix: if str: add the provided string after the concatenated words. if tuple the first string is for singular
    and the second string is for plural.
    """

    def format_noun(noun: StrOrTupleStr, num: int):
        if isinstance(noun, str):
            pass
        elif isinstance(noun, tuple):
            if num_words == 1:
                noun = noun[0]
            else:
                noun = noun[1]
        else:
            raise ValueError(f'prefix must be either str or tuple, not {type(prefix)}')
        if '{}' in noun:
            noun = noun.format(num)
        if '[s]' in noun:
            noun = noun.replace('[s]', 's' if num_words > 1 else '')
        return noun

    # wrap each word with the provided string:
    if isinstance(wrap_with, str):
        words = [wrap_with + str(word) + wrap_with for word in words]
    elif isinstance(wrap_with, tuple):
        words = [wrap_with[0] + str(word) + wrap_with[1] for word in words]
    elif wrap_with is not None:
        raise ValueError(f'wrap_with must be either str or tuple, not {type(wrap_with)}')

    num_words = len(words)

    # concatenate the words:
    last_separator = last_separator or separator
    if num_words == 0:
        return empty_str
    elif num_words == 1:
        s = words[0]
    elif num_words == 2:
        s = words[0] + last_separator + words[1]
    else:
        s = separator.join(words[:-1]) + last_separator + words[-1]

    return format_noun(prefix, num_words) + s + format_noun(suffix, num_words)


class NiceList(list):
    """
    A list that can be printed nicely.
    """
    def __init__(self, *args, wrap_with: StrOrTupleStr = '', prefix: StrOrTupleStr = '',
                 suffix: StrOrTupleStr = '', separator: str = ', ', last_separator: Optional[str] = None,
                 empty_str: str = ''):
        super().__init__(*args)
        self.wrap_with = wrap_with
        self.prefix = prefix
        self.suffix = suffix
        self.separator = separator
        self.last_separator = last_separator
        self.empty_str = empty_str

    def __str__(self):
        return nicely_join(self, self.wrap_with, self.prefix, self.suffix, self.separator, self.last_separator,
                           )

    def __repr__(self):
        return str(self)


class NiceDict(dict):
    """
    A dict that is printed with a new line for each key.
    For example:
    {'key1': value1, 'key2': value2}
    is printed as:
    {
    'key1': value1,
    'key2': value2,
    }
    """
    def __init__(self, *args, format_numerics_and_iterables: Optional[callable] = None):
        super().__init__(*args)
        self.format_numerics_and_iterables = format_numerics_and_iterables or repr

    def __str__(self):
        if len(self) == 0:
            return '{}'
        return '{\n' + '\n'.join([f'    {repr(key)}: {self.format_numerics_and_iterables(value)},'
                                  for key, value in self.items()]) + '\n}'

    def __repr__(self):
        return str(self)



from contextlib import contextmanager
from typing import Optional

import colorama
from functools import partial

from pathlib import Path

from .console_log_to_html import convert_console_log_to_html
from data_to_paper.text.highlighted_text import colored_text
from .mutable import Mutable, Flag

CONSOLE_LOG_FILE = Mutable(None)

IS_LOGGING_ENABLED = Flag(True)


@contextmanager
def console_log_file_context(file_path: Path):
    """
    Context manager to temporarily change the console log file.
    If run is successful, also converts the console log to html.
    """
    global CONSOLE_LOG_FILE
    old_val = CONSOLE_LOG_FILE.val
    CONSOLE_LOG_FILE.val = file_path
    try:
        yield
    except Exception:
        raise
    else:
        convert_console_log_to_html(CONSOLE_LOG_FILE.val)
    finally:
        CONSOLE_LOG_FILE.val = old_val


def print_and_log(text_in_bw: str, text_in_color: Optional[str] = None, color: Optional[str] = None,
                  should_log: bool = True, **kwargs):
    if not IS_LOGGING_ENABLED:
        return
    if color is not None:
        assert text_in_color is None
        text_in_color = colored_text(text_in_bw, color)
    else:
        if text_in_color is None:
            text_in_color = text_in_bw
    print(text_in_color, **kwargs)
    if should_log and CONSOLE_LOG_FILE.val is not None:
        file_path_color = CONSOLE_LOG_FILE.val  # pathlib.Path
        with open(file_path_color, 'a', encoding='utf-8') as f:
            print(text_in_color, file=f, **kwargs)
        file_path_bw = file_path_color.with_stem(file_path_color.stem + '_bw')
        with open(file_path_bw, 'a', encoding='utf-8') as f:
            print(text_in_bw, file=f, **kwargs)


print_and_log_red = partial(print_and_log, color=colorama.Fore.RED)
print_and_log_magenta = partial(print_and_log, color=colorama.Fore.MAGENTA)



from typing import Optional, Tuple

THEME_NAME = 'big_bang_theory'


MAX_BARS: int = 50


KIND_TO_MAX_ROWS_AND_COLUMNS_FOR_PLOT = {
    None: (50, 10),  # Default
    'bar': (50, 10),
    'hist': (None, 10),
}


KIND_TO_MAX_ROWS_AND_COLUMNS_FOR_SHOW = {
    None: (50, 10),  # Default
    'bar': (50, 10),
    'hist': (5, 10),
}


TABLE_MAX_ROWS_AND_COLUMNS: Tuple[Optional[int], Optional[int]] = (20, 6)


def get_max_rows_and_columns(is_figure: bool, kind: Optional[str], to_show: bool = False
                             ) -> tuple[Optional[int], Optional[int]]:
    if not is_figure:
        return TABLE_MAX_ROWS_AND_COLUMNS
    dict_ = KIND_TO_MAX_ROWS_AND_COLUMNS_FOR_SHOW if to_show else KIND_TO_MAX_ROWS_AND_COLUMNS_FOR_PLOT
    return dict_.get(kind, dict_[None])



import pandas as pd

from data_to_paper.llm_coding_utils import df_to_figure
from data_to_paper.research_types.hypothesis_testing.check_df_to_funcs.df_checker import \
    raise_issue_if_filename_not_legit_df_tag
from data_to_paper.run_gpt_code.overrides.dataframes.df_with_attrs import save_as_func_call_df


def analysis_df_to_figure(df: pd.DataFrame, filename: str, **kwargs):
    """
    Replacement of df_to_figure to be used by LLM-writen code.
    We are not actually running df_to_figure (yet).
    Instead, we are just saving the function call.
    Tests and runs are done after the LLM-code is completed.
    """
    raise_issue_if_filename_not_legit_df_tag(filename)
    return save_as_func_call_df(df_to_figure, df, filename, kwargs)



from __future__ import annotations
from pathlib import Path

from typing import Union

from .actions_and_conversations import Actions


def replay_actions(file_path: Union[str, Path], is_color: bool = True):
    """
    Replay a list of actions on conversations.
    """
    actions = Actions()

    actions.load_actions_from_file(file_path)
    for action in actions:
        actions.apply_action(action, is_color=is_color, should_append=False)

    return actions



# Utils made available to the LLM code for the data analysis step
from .df_to_figure import analysis_df_to_figure as df_to_figure
from .df_to_latex import analysis_df_to_latex as df_to_latex
df_to_latex.__qualname__ = 'df_to_latex'
df_to_figure.__qualname__ = 'df_to_figure'



from abc import ABCMeta
from dataclasses import dataclass
from typing import Optional, Union

from data_to_paper.conversation.stage import Stage
from data_to_paper.exceptions import data_to_paperException


class TerminateException(data_to_paperException, metaclass=ABCMeta):
    """
    Base class for all exceptions that terminate data-to-paper run.
    """
    pass


@dataclass
class MissingInstallationError(TerminateException):
    """
    An exception to indicate a missing installation.
    """
    package_name: str
    instructions: Optional[str] = None

    def __str__(self):
        s = f"{self.package_name} is not installed. Please install it."
        if self.instructions:
            s += f"\n\n{self.instructions.strip()}"
        return s


@dataclass
class ResetStepException(Exception):
    """
    An exception to reset the step.
    """
    stage: Optional[Union[Stage, bool]]  # reset to this stage, None to exit the app



from pandas import DataFrame


def get_original_df_method(method_name):
    """
    Get the original DataFrame method.
    """
    method = getattr(DataFrame, method_name)
    while hasattr(method, 'wrapper_of'):
        method = method.wrapper_of
    return method



from copy import copy

import numpy as np

from dataclasses import dataclass, field
from typing import Optional, Dict, List, Iterable, NamedTuple

from data_to_paper.base_products.product import ValueProduct
from data_to_paper.code_and_output_files.file_view_params import ViewPurpose
from data_to_paper.env import SCHOLAR_SERVER
from data_to_paper.utils.iterators import interleave
from data_to_paper.utils.mutable import Flag
from data_to_paper.utils.nice_list import NiceList
from data_to_paper.servers.custom_types import Citation


def unite_citation_lists(
    citations_lists: Iterable[Iterable[Citation]], total: int = None
) -> List[Citation]:
    """
    Unite the two lists maintaining a single bibtex_id for each citation.
    Citation that appears in both list will have the query as a set of both quereis.
    """
    bibtex_ids_to_citations = {}
    for citation in interleave(*citations_lists):
        bibtex_id = citation.bibtex_id
        query = citation.query
        if isinstance(query, str):
            query = {query}
        if bibtex_id not in bibtex_ids_to_citations:
            citation = copy(citation)
            citation.query = query
            bibtex_ids_to_citations[bibtex_id] = citation
        else:
            bibtex_ids_to_citations[bibtex_id].query.update(query)
            bibtex_ids_to_citations[bibtex_id].search_rank = min(
                bibtex_ids_to_citations[bibtex_id].search_rank, citation.search_rank
            )
        if total is not None:
            if len(bibtex_ids_to_citations) == total:
                break
    return list(bibtex_ids_to_citations.values())


CITATION_REPR_FIELDS_FOR_LLM = (
    "bibtex_id",
    "title",
    "journal_and_year",
    "tldr",
    "influence",
)
CITATION_REPR_FIELDS_FOR_PRINT = (
    "query",
    "search_rank",
    "bibtex_id",
    "title",
    "journal_and_year",
    "tldr",
    "influence",
    "embedding_similarity",
)
GET_LITERATURE_SEARCH_FOR_PRINT = Flag(False)


@dataclass
class LiteratureSearchQueriesProduct(ValueProduct):
    name: str = "Literature Search Queries"
    value: Dict[str, List[str]] = None

    def _get_content_as_formatted_text(
        self, level: int, view_purpose: ViewPurpose, **kwargs
    ) -> str:
        s = ""
        level_str = "#" * (level + 1)
        for scope, queries in self.value.items():
            s += f"{level_str} {scope.title()}\n"
            for query in queries:
                s += f'- "{query}"\n'
        return s


@dataclass
class CitationCollectionProduct(ValueProduct):
    name: str = "Citation List"
    value: List[Citation] = None

    def __iter__(self) -> Iterable[Citation]:
        return super().__iter__()

    def _get_citations_as_str(
        self,
        is_html: bool,
        style: str = None,
        embedding_target: Optional[np.ndarray] = None,
    ) -> str:
        return "\n".join(
            citation.pretty_repr(
                fields=(
                    CITATION_REPR_FIELDS_FOR_LLM
                    if style == "llm"
                    else CITATION_REPR_FIELDS_FOR_PRINT
                ),
                is_html=is_html,
                embedding_target=embedding_target,
            )
            for citation in self.value
        )

    def _get_content_as_formatted_text(
        self,
        level: int,
        view_purpose: ViewPurpose,
        style: str = None,
        embedding_target: Optional[np.ndarray] = None,
        **kwargs: str,
    ) -> str:
        return self._get_citations_as_str(
            is_html=False, style=style, embedding_target=embedding_target
        )

    def _get_content_as_html(
        self,
        level: int,
        style: str = None,
        embedding_target: Optional[np.ndarray] = None,
    ):
        return self._get_citations_as_str(is_html=True, style=style)


@dataclass
class QueryCitationCollectionProduct(CitationCollectionProduct):
    name: str = "Query Citation List"
    query: str = None

    def get_header(
        self, view_purpose: ViewPurpose = ViewPurpose.PRODUCT, **kwargs
    ) -> str:
        return f'{self.name} for "{self.query}"'


@dataclass
class SortedCitationCollectionProduct(CitationCollectionProduct):
    name: str = "Sorted Citation List"
    scope: Optional[str] = None
    query: Optional[str] = None
    total: int = None
    distribution_factor: Optional[float] = None
    sort_by_similarity: bool = False
    minimal_influence: int = 0

    def get_header(
        self,
        view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
        style: str = "llm",
        **kwargs,
    ) -> str:
        s = super().get_header()
        s += f', Scope: "{self.scope}"' if self.scope is not None else ""
        s += f', Query: "{self.query}"' if self.query is not None else ""
        if style != "llm":
            s += f", Total: {self.total}" if self.total is not None else ""
            s += (
                f", Distribution Factor: {self.distribution_factor}"
                if self.distribution_factor is not None
                else ""
            )
            s += (
                f", Sort by Similarity: {self.sort_by_similarity}"
                if self.sort_by_similarity
                else ""
            )
            s += (
                f", Minimal Influence: {self.minimal_influence}"
                if self.minimal_influence > 0
                else ""
            )
        return s


class LiteratureSearchParams(NamedTuple):
    total: int
    minimal_influence: int
    distribution_factor: Optional[float]
    sort_by_similarity: bool

    def to_dict(self) -> dict:
        return self._asdict()


@dataclass
class LiteratureSearch(ValueProduct):
    # value is scopes_to_queries_to_citations
    value: Dict[str, Dict[str, CitationCollectionProduct]] = field(default_factory=dict)
    embedding_target: Optional[np.ndarray] = None
    scopes_to_search_params: Dict[str, LiteratureSearchParams] = field(
        default_factory=dict
    )
    api_source: str = SCHOLAR_SERVER.server

    def get_queries(self, scope: Optional[str] = None) -> List[str]:
        """
        Return the queries in the given scope.
        if scope=None, return all queries.
        """
        if scope is None:
            queries = sum([self.get_queries(scope) for scope in self], [])
        else:
            queries = list(self[scope].keys())
        return NiceList(
            queries, wrap_with='"', separator="\n", prefix="\n", suffix="\n"
        )

    def get_citations(
        self,
        scope: Optional[str] = None,
        query: Optional[str] = None,
        total: int = None,
        distribution_factor: Optional[float] = None,
        sort_by_similarity: bool = False,
        minimal_influence: int = 0,
    ) -> CitationCollectionProduct:
        """
        Return the citations in the given scope.
        If embedding_target is not None, sort the citations by embedding similarity.
        If total is not None, return only the first total citations.
        If total < 0, return only the last total citations.
        """
        if scope is None:
            assert query is None
            citations = unite_citation_lists(
                (
                    iter(
                        self.get_citations(
                            scope=scope,
                            total=(
                                int(total / len(self) * distribution_factor) + 1
                                if distribution_factor and total is not None
                                else None
                            ),
                            minimal_influence=minimal_influence,
                            distribution_factor=distribution_factor,
                            sort_by_similarity=sort_by_similarity,
                        )
                    )
                    for scope in self
                )
            )
        elif query is None:
            citations = unite_citation_lists(
                (
                    iter(
                        self.get_citations(
                            scope=scope,
                            query=query,
                            total=(
                                int(total / len(self[scope]) * distribution_factor) + 1
                                if distribution_factor and total is not None
                                else None
                            ),
                            minimal_influence=minimal_influence,
                            distribution_factor=distribution_factor,
                            sort_by_similarity=sort_by_similarity,
                        )
                    )
                    for query in self[scope]
                )
            )
        else:
            citations = self[scope][query]
        citations = list(citations)

        if self.api_source == "Crossref":
            # Remove citations without title and abstract
            citations = [
                citation
                for citation in citations
                if (citation.title is not None and citation.title != "")
                and (citation.abstract is not None and citation.abstract != "")
            ]

        else:
            if minimal_influence > 0:
                citations = [
                    citation
                    for citation in citations
                    if citation.influence >= minimal_influence
                ]

            if sort_by_similarity and self.embedding_target is not None:
                citations = sorted(
                    citations,
                    key=lambda citation: citation.get_embedding_similarity(
                        self.embedding_target
                    ),
                    reverse=True,
                )
            else:
                citations = sorted(citations, key=lambda citation: citation.search_rank)

        if total is None:
            pass
        elif total < 0:
            citations = citations[total:]
        else:
            citations = citations[:total]
        return SortedCitationCollectionProduct(
            value=citations,
            scope=scope,
            query=query,
            total=total,
            distribution_factor=distribution_factor,
            sort_by_similarity=sort_by_similarity,
            minimal_influence=minimal_influence,
        )

    def pretty_repr(
        self,
        with_scope_and_queries: bool = False,
        total: int = None,
        distribution_factor: Optional[float] = None,
        sort_by_similarity: bool = False,
        minimal_influence: int = 0,
        style: str = None,
    ) -> str:
        s = ""
        for scope in self:
            if with_scope_and_queries:
                s += "\n\n"
                s += f"Scope: {repr(scope)}\n"
                s += f"Queries: {repr(self.get_queries(scope))}\n"
            s += self.pretty_repr_for_scope_and_query(
                scope=scope,
                total=total // len(self) + 1,
                distribution_factor=distribution_factor,
                sort_by_similarity=sort_by_similarity,
                minimal_influence=minimal_influence,
                style=style,
            )
        return s

    def pretty_repr_for_scope_and_query(
        self,
        scope: str,
        query: Optional[str] = None,
        total: int = None,
        distribution_factor: Optional[float] = None,
        sort_by_similarity: bool = False,
        minimal_influence: int = 0,
        style: str = None,
    ) -> str:
        """
        style: 'llm', 'print', 'html'
        """
        style = style or "llm"
        citations = self.get_citations(
            scope=scope,
            query=query,
            total=total,
            distribution_factor=distribution_factor,
            sort_by_similarity=sort_by_similarity,
            minimal_influence=minimal_influence,
        )
        if GET_LITERATURE_SEARCH_FOR_PRINT:
            style = "print"
        return "\n".join(
            citation.pretty_repr(
                fields=(
                    CITATION_REPR_FIELDS_FOR_LLM
                    if style == "llm"
                    else CITATION_REPR_FIELDS_FOR_PRINT
                ),
                is_html=style == "html",
                embedding_target=self.embedding_target,
            )
            for citation in citations
        )

    def get_header(
        self,
        view_purpose: ViewPurpose = ViewPurpose.PRODUCT,
        scope: Optional[str] = None,
        **kwargs,
    ) -> str:
        return f"{scope}-related literature search" if scope is not None else self.name

    def _get_content_as_formatted_text(
        self,
        level: int,
        view_purpose: ViewPurpose,
        scope: Optional[str] = None,
        style: str = "llm",
        **kwargs,
    ) -> str:
        if scope is None:
            s = ""
            for scope in self:
                s += self._get_content_as_formatted_text(
                    level, view_purpose, scope=scope, style=style
                )
        else:
            s = self.pretty_repr_for_scope_and_query(
                scope=scope,
                style=style,
                **self.scopes_to_search_params[scope].to_dict(),
            )
        return s

    def _get_content_as_html(
        self, level: int, scope: Optional[str] = None, style: str = "html", **kwargs
    ) -> str:
        if scope is None:
            s = "We searched for papers in the following scopes:\n"
            for scope in self:
                s += f"<h{level + 1}>{scope.title()}-related papers</h{level + 1}>\n"
                s += self._get_content_as_html(level, scope=scope, style=style)
        else:
            s = self.pretty_repr_for_scope_and_query(
                scope=scope,
                style=style,
                **self.scopes_to_search_params[scope].to_dict(),
            )
        return s



from dataclasses import dataclass
from typing import Tuple, Optional, Dict, Any

from data_to_paper.code_and_output_files.output_file_requirements import OutputFileRequirements, \
    DataOutputFileRequirement
from data_to_paper.research_types.hypothesis_testing.coding.base_code_conversers import BaseScientificCodeProductsGPT
from data_to_paper.research_types.hypothesis_testing.coding.utils import create_pandas_and_stats_contexts
from data_to_paper.text import dedent_triple_quote_str


@dataclass
class DataPreprocessingCodeProductsGPT(BaseScientificCodeProductsGPT):
    code_step: str = 'data_preprocessing'
    background_product_fields: Tuple[str, ...] = ('research_goal', 'all_file_descriptions', 'outputs:data_exploration')
    # user_agent: ScientificAgent = ScientificAgent.DataPreprocessor
    allow_data_files_from_sections: Tuple[Optional[str]] = (None, 'data_exploration', )
    supported_packages: Tuple[str, ...] = ('pandas', 'numpy', 'scipy', 'imblearn')

    mission_prompt: str = dedent_triple_quote_str("""
        As part of a data-preprocessing phase, please write a complete short Python code for getting a \t
        cleaned, normalized, same-unit, balanced version of the data, ready for use in following analysis \t
        steps that will include statistical tests and/or machine learning models on the processed data.

        Your code should create one or more new csv files containing the preprocessed data, saved with \t
        sensible file names.

        Depending on the specifics of the dataset and the goal and hypothesis specified above, \t
        you might want to preform the following steps:

        * Dealing with missing values - imputation, deletion, etc.
        * Normalization of numeric values with different units into same-unit values.
        * Scaling numeric values into a common scale (e.g., 0-1) using min-max scaling, z-score, etc.
        * Encoding categorical variables into numeric values (e.g., using one-hot encoding)
        * Balancing the data by under-sampling, over-sampling, or more advanced techniques to deal with class imbalance
        * Any other data preprocessing you deem relevant

        You are not obliged to perform all of the above steps, choose the ones that suits the data and the hypothesis
        we are testing (see research goal above). 

        If needed, you can use the following packages which are already installed:
        {supported_packages}

        Do not provide a sketch or pseudocode; write a complete runnable code.
        Do not create any graphics, figures or any plots.
        """)

    def _create_output_file_requirements(self) -> OutputFileRequirements:
        return OutputFileRequirements([DataOutputFileRequirement('*.csv')])

    def _get_additional_contexts(self) -> Dict[str, Any]:
        return create_pandas_and_stats_contexts(allow_dataframes_to_change_existing_series=False,
                                                enforce_saving_altered_dataframes=True)



from .methods import __init__, to_string, to_csv, __str__, __setitem__, __getitem__, __delitem__, \
    __LocationIndexer__get_item__

from .methods import STR_FLOAT_FORMAT

from .raise_on_call import raise_on_call



from __future__ import annotations

import os
from contextlib import contextmanager, ExitStack
from dataclasses import dataclass
from typing import List, Type, TypeVar, Optional, Iterable, Union, Dict

from .user_script_name import is_called_from_user_script
from .run_issues import RunIssues

T = TypeVar('T', bound='SingletonRegisteredRunContext')


@dataclass
class DisableableContext:
    """
    Context manager that can be temporarily disabled.
    """
    _is_enabled: Optional[bool] = None

    def _reversible_enter(self):
        self._is_enabled = True

    def _reversible_exit(self):
        self._is_enabled = False

    def __enter__(self):
        self._reversible_enter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._reversible_exit()
        self._is_enabled = None
        return False

    @contextmanager
    def temporarily_disable(self):

        assert self._is_enabled is not None, 'The context is not initialized yet.'

        was_enabled = self._is_enabled
        assert was_enabled is not None, 'The context is not initialized yet.'
        if was_enabled:
            self._reversible_exit()
        try:
            yield
        finally:
            if was_enabled:
                self._reversible_enter()


@dataclass
class RunContext(DisableableContext):
    """
    Base context manager for running GPT code.
    """
    issues: Optional[RunIssues] = None
    name: Optional[str] = None

    def __enter__(self):
        self.issues = RunIssues()
        return super().__enter__()

    def _is_called_from_user_script(self, offset: int = 3) -> bool:
        """
        Check if the code is called from user script.
        """
        return is_called_from_user_script(offset=offset + 1)


@dataclass
class RegisteredRunContext(RunContext):
    """
    Base context manager for running GPT code that can be registered to be accessed from anywhere.
    Allows centralized `temporarily_disable_all` registered context managers.
    """
    PROCESS_AND_NAME_TO_OBJECT = {}
    should_register: bool = True

    @property
    def _process_and_identifier(self):
        return os.getpid(), self._identifier

    @property
    def _identifier(self):
        return id(self)

    def __enter__(self):
        if self.should_register:
            self.PROCESS_AND_NAME_TO_OBJECT[self._process_and_identifier] = self
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.should_register:
            self.PROCESS_AND_NAME_TO_OBJECT.pop(self._process_and_identifier)
        return super().__exit__(exc_type, exc_val, exc_tb)

    @classmethod
    def get_all_runtime_instances(cls) -> List[RegisteredRunContext]:
        process_id = os.getpid()
        return [obj for (pid, _), obj in cls.PROCESS_AND_NAME_TO_OBJECT.items() if pid == process_id]

    @staticmethod
    @contextmanager
    def temporarily_disable_all():
        """
        Context manager for temporarily disabling all registered runtime contexts.
        """
        objs = RegisteredRunContext.get_all_runtime_instances()
        with ExitStack() as stack:
            for obj in objs:
                stack.enter_context(obj.temporarily_disable())
            yield


@dataclass
class SingletonRegisteredRunContext(RegisteredRunContext):
    """
    Base context manager for running GPT code that only exists once per class.
    Allows accessing the runtime instance from the class.
    """

    @property
    def _identifier(self):
        return self.__class__.__name__

    def __enter__(self):
        if self._identifier in self.PROCESS_AND_NAME_TO_OBJECT:
            raise RuntimeError(f'SingletonRegisteredRunContext {self._identifier} already exists.')
        return super().__enter__()

    @classmethod
    def get_runtime_instance(cls: Type[T]) -> T:
        process_and_identifier = os.getpid(), cls.__name__
        if process_and_identifier not in cls.PROCESS_AND_NAME_TO_OBJECT:
            raise RuntimeError(f'SingletonRegisteredRunContext {cls.__name__} was not created yet.')
        return cls.PROCESS_AND_NAME_TO_OBJECT[process_and_identifier]


@dataclass
class MultiRunContext(RunContext):
    """
    Pack together multiple run contexts.
    """
    contexts: Union[Dict[str, RunContext], Iterable[RunContext]] = ()
    exit_stack: ExitStack = None

    def get_contexts(self) -> List[RunContext]:
        if isinstance(self.contexts, dict):
            return list(self.contexts.values())
        return list(self.contexts)

    def __enter__(self):
        self.exit_stack = ExitStack()
        for context in self.get_contexts():
            self.exit_stack.enter_context(context)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        for context in self.get_contexts():
            self.issues.extend(context.issues)
        super_result = super().__exit__(exc_type, exc_val, exc_tb)
        stack_result = self.exit_stack.__exit__(exc_type, exc_val, exc_tb)
        return super_result or stack_result



import time
from typing import List, Mapping, Any

import requests
from unidecode import unidecode

from data_to_paper.utils.nice_list import NiceList

from .base_server import ParameterizedQueryServerCaller
from .custom_types import Citation
from .types import ServerErrorException

from data_to_paper.utils.print_to_file import print_and_log_red

CROSSREF_URL = "https://api.crossref.org/works"

HEADERS = {"User-Agent": "data_to_paper (mailto:kishonystud@technion.ac.il)"}

BIBTEX_TEMPLATE = "@{type}{{{id},\n{fields}}}\n"

TYPE_MAPPING = {
    "journal-article": "article",
    "proceedings-article": "inproceedings",
    "book": "book",
    "edited-book": "inbook",
    "book-chapter": "inbook",
}


def get_type_from_crossref(crossref: Mapping[str, Any]) -> str:
    return TYPE_MAPPING.get(crossref["type"], "misc")


PAPER_MAPPING = {
    "authors": "author",
    "title": "title",
    "journal": "journal",
    "year": "year",
    "volume": "volume",
    "issue": "number",
    "pages": "pages",
    "DOI": "doi",
}

BOOK_MAPPING = {
    "authors": "author",
    "editors": "editor",
    "title": "title",
    "publisher": "publisher",
    "year": "year",
    "volume": "volume",
    "issue": "number",
    "pages": "pages",
    "DOI": "doi",
    "ISBN": "isbn",
}


STRS_TO_REMOVE_FROM_BIBTEX_ID = [
    " ",
    "-",
    "_",
    "–",
    "’",
    "'",
    "/",
    " ",
    "(",
    ")",
    "[",
    "]",
    "{",
    "}",
    ":",
    ";",
    ",",
    ".",
    "?",
    "!",
    "“",
    "”",
    '"',
    "–",
]

FORBIDDEN_BIBTEX_IDS = set(
    [
        "None",
        "none",
        "nan",
        "NA",
        "na",
        "N/A",
        "Introduction",
        "introduction",
        "Results",
        "results",
        "Discussion",
        "discussion",
        "Methods",
        "methods",
        "Abstract",
        "abstract",
        "Conclusion",
        "conclusion",
        "Background",
        "background",
        "Acknowledgments",
        "acknowledgments",
        "References",
        "references",
        "Supplementary",
        "supplementary",
        "",
        " ",
    ]
)


class CrossrefCitation(Citation):
    """
    A single crossref citation. This the raw dict after transforming the json response.
    This dict is hashable.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def bibtex_type(self) -> str:
        return get_type_from_crossref(self)

    @property
    def bibtex(self) -> str:
        # create a mapping for article and inproceedings
        bibtex_type = get_type_from_crossref(self)
        is_paper = bibtex_type in ["article", "inproceedings"]
        field_mapping = PAPER_MAPPING if is_paper else BOOK_MAPPING

        fields = []
        for key, value in self.items():
            if key in field_mapping:
                if value and key not in ["doi", "isbn"]:
                    # remove special characters of the value
                    if isinstance(value, list):
                        value = [
                            unidecode(v).replace(r" &", r" \&").replace(r"_", r"\_")
                            for v in value
                        ]
                    elif isinstance(value, str):
                        value = (
                            unidecode(value).replace(r" &", r" \&").replace(r"_", r"\_")
                        )
                bibtex_key = field_mapping[key]
                fields.append(f"{bibtex_key} = {{{value}}}")
        return BIBTEX_TEMPLATE.format(
            type=self.bibtex_type, id=self.bibtex_id, fields=",\n".join(fields)
        )

    @property
    def bibtex_id(self) -> str:
        """
        Get the bibtex id for this citation.
        """
        bibtex_id = unidecode(self["first_author_family"]) + (
            str(self.get("year")) if self.get("year") else ""
        )
        bibtex_id += self["title"].split(" ")[0] if self.get("title") else ""

        # remove special characters from the id like -, _, etc.
        for char in STRS_TO_REMOVE_FROM_BIBTEX_ID:
            bibtex_id = bibtex_id.replace(char, "")
        return bibtex_id

    @property
    def title(self) -> str:
        """
        Get the title of the citation.
        """
        return self.get("title", "")

    @property
    def abstract(self) -> str:
        """
        Get the abstract of the citation.
        """
        return self.get("abstract", "")

    @property
    def journal(self) -> str:
        """
        Get the journal of the citation.
        """
        return self.get("journal", "")

    @property
    def year(self) -> str:
        """
        Get the year of the citation.
        """
        return self.get("year", "")

    @property
    def influence(self) -> int:
        """
        Get the influence of the citation.
        """
        return self.get("influence", 0)

    @property
    def embedding(self) -> str:
        """
        Get the embedding of the citation.
        """
        return self.get("embedding", None)

    @property
    def tldr(self) -> str:
        """
        TLDR will be the abstract if exists, otherwise the title, otherwise None.
        """
        abstract = self.get("abstract", None)
        if abstract:
            return abstract
        title = self.get("title", None)
        if title:
            return title
        return None


class CrossrefServerCaller(ParameterizedQueryServerCaller):
    """
    Search for citations in Crossref.
    """

    name = "Crossref"
    file_extension = "_crossref.bin"

    @staticmethod
    def crossref_item_to_citation(item: dict) -> dict:
        # create authors as a string in the format
        # "first_name1 last_name1 and first_name2 last_name2 and first_name3 last_name3"
        authors_string = ""
        if item.get("author"):
            for author in item["author"]:
                if author != item["author"][-1]:
                    authors_string += (
                        f"{author.get('given', '')} {author.get('family', '')} and "
                    )
                else:
                    authors_string += (
                        f"{author.get('given', '')} {author.get('family', '')}"
                    )
        first_author = item.get("author", [None])[0]
        first_author_family = first_author.get("family", "") if first_author else ""
        citation = {
            "title": item.get("title", [""])[0],
            "abstract": item.get("abstract", ""),
            "first_author_family": first_author_family,
            "authors": authors_string,
            "journal": item.get("container-title", [None])[0],
            "doi": item.get("DOI", ""),
            "type": item.get("type", ""),
            "year": (
                item["published"]["date-parts"][0][0]
                if "published" in item
                else (
                    item["published-print"]["date-parts"][0][0]
                    if "published-print" in item
                    else ""
                )
            ),
            "publisher": item.get("publisher", ""),
            "volume": item.get("volume", ""),
            "issue": item.get("issue", ""),
            "page": item.get("page", ""),
            "isbn": item.get("ISBN", ""),
        }
        for key, value in citation.items():
            if isinstance(value, str) and "&NA;" in value:
                citation[key] = ""
        return citation

    @classmethod
    def _get_server_response(cls, query, rows=4) -> List[dict]:
        """
        Get the response from the crossref server as a list of CrossrefCitation objects.
        """

        params = {
            "query.bibliographic": query,
            "rows": rows,
            "filter": "type:journal-article,type:proceedings-article,type:book,type:edited-book,type:book-chapter",
            "select": "title,author,container-title,DOI,type,published,published-print,publisher,volume,issue,page,"
            "editor,ISBN",
        }
        print_and_log_red(f'QUERYING Crossref FOR: "{query}"', should_log=False)
        for attempt in range(3):
            response = requests.get(CROSSREF_URL, headers=HEADERS, params=params)
            if response.status_code not in (504, 429):
                break
            wait_time = 2**attempt  # Exponential backoff
            print_and_log_red(
                f"ERROR: Server timed out or too many requests. "
                f"We wait for {wait_time} sec and try again.",
                should_log=False,
            )
            time.sleep(wait_time)
        else:
            raise ServerErrorException(
                server=cls.name, response=response
            )  # if we failed all attempts

        if response.status_code != 200:  # 200 is the success code
            raise ServerErrorException(server=cls.name, response=response)

        data = response.json()
        papers = [
            cls.crossref_item_to_citation(item) for item in data["message"]["items"]
        ]

        if len(papers) > 0:  # if there are papers
            papers = [
                paper for paper in papers if CrossrefCitation(paper).bibtex_id != "None"
            ]
            return papers
        else:
            # failing gracefully
            return []

    @staticmethod
    def _post_process_response(
        response: List[dict], args, kwargs
    ) -> List[CrossrefCitation]:
        """
        Post process the response from the server.
        """
        query = args[0] if args else kwargs.get("query", None)
        citations = NiceList(separator="\n", prefix="[\n", suffix="\n]")
        for rank, paper in enumerate(response):
            citation = CrossrefCitation(paper, search_rank=rank, query=query)
            if (
                len(citation.bibtex_id) <= 4
                or citation.bibtex_id in FORBIDDEN_BIBTEX_IDS
            ):
                # bibtex_id is too short or is forbidden
                print_and_log_red(
                    f"ERROR: bibtex_id is too short. skipping. Title: {citation.title}"
                )
                continue
            citations.append(citation)
        return citations


CROSSREF_SERVER_CALLER = CrossrefServerCaller()



"""
Base classes to use for building each step in a multi-step process towards a goal.
"""

# --- PRODUCTS ---

# Products:

# Basic Products types:

# --- RUNNING MULTI-STEP PROCESS ---

# Base class for running multiple steps while accumulating Products towards a goal:
from .base_steps_runner import BaseStepsRunner, DataStepRunner

# In each step, we can use the Products from the previous step and choose
# from one of the base-classes below to create new Products.

# --- REQUESTING PRODUCTS FROM USER ---

# Base classes for requesting the user for products:
from .request_products_from_user import DirectorProductGPT


# --- REQUESTS PRODUCTS FROM LLM ---

# Requesting un-structured text:
from .base_products_conversers import BackgroundProductsConverser

# Requesting un-structured text as part of a gpt-gpt review process:
from .base_products_conversers import ReviewBackgroundProductsConverser

# Addon to check if response correctly extracts from provided products:
from .base_products_conversers import \
    CheckExtractionReviewBackgroundProductsConverser, CheckReferencedNumericReviewBackgroundProductsConverser

# Requesting quote-enclosed text (with optional gpt-review):
from .request_quoted_text import BaseProductsQuotedReviewGPT

# Requesting LaTeX formatted text (with optional gpt-review):
from .request_latex import LatexReviewBackgroundProductsConverser

# Requesting Python values (with optional gpt-review):
from .request_python_value import PythonValueReviewBackgroundProductsConverser
from .request_python_value import PythonDictReviewBackgroundProductsConverser
from .request_python_value import PythonDictWithDefinedKeysReviewBackgroundProductsConverser
from .request_python_value import PythonDictWithDefinedKeysAndValuesReviewBackgroundProductsConverser

# Requesting code (with automatic debugging feedback):
from .request_code import BaseCodeProductsGPT
from .debugger import DebuggerConverser

# Requesting answer to multiple choice question
from .request_multi_choice import MultiChoiceBackgroundProductsConverser

# Requesting literature search
from .request_literature_search import BaseLiteratureSearchReviewGPT
from .literature_search import LiteratureSearch

# --- CONVERTING PRODUCTS TO FILES ---

# Base classes for creating files from Products:
from .base_products_to_file import BaseFileProducer

# Base classes for creating PDFs from LaTeX Products:
from .latex_products_to_pdf import BaseLatexToPDF



import os
import sys


def open_file_on_os(file_path):
    """
    Open a file with the default application.
    """
    if sys.platform == 'win32':
        os.startfile(file_path)
    else:
        # For macOS and Linux, we use 'open' and 'xdg-open' respectively
        opener = 'open' if sys.platform == 'darwin' else 'xdg-open'
        os.system(f'{opener} "{file_path}"')



from enum import Enum
from typing import NamedTuple, Optional


class Algorithm(Enum):
    GPT = 'GPT'
    PRE_PROGRAMMED = 'PRE_PROGRAMMED'
    MIXED = 'MIXED'

    def pretty_repr(self, prompt: Optional[str] = None) -> str:
        if prompt is None:
            prompt_statement = f"I am still waiting to be given a mission statement (please check in later)."
        else:
            prompt_statement = f"I have been given the following mission statement:\n{prompt}\n"

        if self == Algorithm.GPT:
            return f"I am run by GPT.\n{prompt_statement}"
        elif self == Algorithm.PRE_PROGRAMMED:
            return f"I am run by a pre-programmed algorithm issuing automated programmed responses."
        elif self == Algorithm.MIXED:
            return f"I am run by combination of a pre-programmed algorithm issuing automated programmed responses,\n" \
                   f"and by GPT.\n{prompt_statement}"
        else:
            raise ValueError(f"Unknown algorithm: {self}")


class Profile(NamedTuple):
    agent_name: str
    name: str
    title: str
    description: str
    image_file: str
    algorithm: Algorithm



# ALLOWED_PLOT_KINDS = ['line', 'scatter', 'bar', 'hist', 'box', 'kde', 'hexbin', 'pie']
import numbers
from typing import Union

from data_to_paper.utils.mutable import Mutable


AXES_SIZE_INCHES = (2.6, 2)  # actual size of the axes box in inches (not including labels, etc.)
FIG_SIZE_INCHES = (4, 3)  # this is not consequential, as we fit the figure to the axes (using fit_fig_to_axes)

FIG_DPI = Mutable(400)

ALLOWED_PLOT_KINDS = ['bar']  # TODO: Add support for more plot kinds

DF_ALLOWED_VALUE_TYPES = (numbers.Number, str, bool, tuple)
DF_ALLOWED_COLUMN_TYPES = (numbers.Number, str, bool)
DfColumnTyping = Union[numbers.Number, str, bool]  # not tuple because we need tuple to choose a pair of columns
